
use std::rc::Rc;
use std::sync::Arc;
use std::thread;
use std::time::{Duration, Instant};

use glow::HasContext;


use bevy_ecs::prelude::*;

use winit::{event_loop::EventLoopBuilder, window::WindowBuilder};
use winit::event::{Event, WindowEvent};

use glutin::{config::ConfigTemplateBuilder, context::ContextAttributesBuilder};
use glutin_winit::DisplayBuilder;
use glutin::{prelude::*, context::Version};
use glutin_winit::{self, GlWindow};
use glutin::display::GetGlDisplay;

pub mod renderer;
fn main() {

    // NOTE initialization code here


    // create Event Loop
    let event_loop = EventLoopBuilder::new().build();

    // set event loop to poll events every frame
    //event_loop.set_control_flow(winit::event_loop::ControlFlow::Poll);


    let window_builder = Some(WindowBuilder::new()
        .with_transparent(true)
    );

    let template = ConfigTemplateBuilder::new();

    let display_builder = DisplayBuilder::new()
        .with_window_builder(window_builder);

    let (window, gl_config) = display_builder
        .build(&event_loop, template, |configs| {
            configs.reduce(|accum, config| {
                let transparency_check = config.supports_transparency().unwrap_or(true)
                    & !accum.supports_transparency().unwrap_or(true);

                if transparency_check || config.num_samples() > accum.num_samples() {
                    config
                }

                else {
                    accum
                }
            })
                   .unwrap()
        })
        .unwrap();


    // probably not needed here
    //let raw_window_handle = window.as_ref().map(|window| window.window_handle()
    //                                            .unwrap()
    //                                            .as_raw());

    let gl_display = gl_config.display();

    let context_attributes = ContextAttributesBuilder::new().build(None);

    let fallback_attributes = ContextAttributesBuilder::new()
        .with_context_api(glutin::context::ContextApi::Gles(None))
        .build(None);

    let compatibility_context_attributes = ContextAttributesBuilder::new()
        .with_context_api(glutin::context::ContextApi::OpenGl(Some(Version::new(3, 3))))
        .build(None);

    // decide on the opengl Context to use, will use the newest one available
    // If the device does NOT at least support opengl 3.3 it will crash
    // opengl 3.3 is supported by most 64 bit computers
    let mut not_current_gl_context = Some(unsafe {
        gl_display.create_context(&gl_config, &context_attributes).unwrap_or_else(|_| {
            gl_display.create_context(&gl_config, &fallback_attributes).unwrap_or_else(|_| {
                gl_display
                    .create_context(&gl_config, &compatibility_context_attributes)
                    .expect("get a newer Computer")
            },
            )
        })
    });


    // current Window Reference Counter
    let window = Rc::new(window.unwrap());

    // window title
    window.set_title("MGQ Fighter Test");

    let attrs = window.build_surface_attributes(<_>::default());

    let gl_surface = unsafe {
        gl_config.display().create_window_surface(&gl_config, &attrs).unwrap()
    };

    let gl_context = not_current_gl_context
        .take()
        .unwrap()
        .make_current(&gl_surface)
        .unwrap();

    // the Glow context wrapped in an Atomic Reference Counter
    let gl = Arc::new(unsafe {
        glow::Context::from_loader_function(|s| {
            let s = std::ffi::CString::new(s).unwrap();
            gl_display.get_proc_address(s.as_c_str()).cast()
        })
    });




    // TODO write bevy Init code here

    let mut world = World::new();
    let mut schedule = Schedule::default();

    schedule.add_systems((
        renderer::test_draw
    ));

    let init_defaults = world.register_system(renderer::initialize_defaults);
    let initialize_test = world.register_system(renderer::test_init);

    world.insert_resource(renderer::GL(gl.clone()));
    let _ = world.run_system(init_defaults);
    let _ = world.run_system(initialize_test);

    let _ = event_loop.run(move |event, _ ,control_flow| {
        // new_frame for Framerate Locking
        let new_frame = Instant::now() + Duration::from_secs_f32(1.0/60.0);
        match event
        {
            Event::WindowEvent { event, .. } => {
                match event
                {
                    WindowEvent::CloseRequested => {
                        println!("closing gracefully");
                        control_flow.set_exit();
                    },
                    WindowEvent::Resized(size) => {
                        if size.width !=0 && size.height !=0 {
                            gl_surface.resize(
                                &gl_context,
                                std::num::NonZeroU32::new(size.width).unwrap(),
                                std::num::NonZeroU32::new(size.height).unwrap()
                            );
                        }
                    },
                    // TODO to be implemented later
                    _ => ()
                };

            }

            Event::Resumed => {

            },
            Event::MainEventsCleared => {
                unsafe {

                    gl.clear_color(0.0, 0.0, 0.0, 0.5);
                    gl.clear(glow::COLOR_BUFFER_BIT);
                }

                schedule.run(&mut world);
                gl_surface.swap_buffers(&gl_context).unwrap();
            },

            // TODO to be implemented
            _ => ()

        }

        // will wait until it's time for the next frame
        thread::sleep(Instant::now() - new_frame);
    });
}
