use glow::HasContext;
use std::{collections::HashMap, path::Path};

use bevy_ecs::prelude::*;

/// SpriteSheet Struct for Representing a Spritesheet
///
/// WARNING while it is very fast to clone the spritesheet, it is assumed that that
/// the spritesheet object is the only one of it's kind, so before it gets dropped make sure to delete
/// the texture from vram
#[derive(Clone)]
pub struct SpriteSheet
{
    pub texture_id: glow::NativeTexture,
    // NOTE where the spritesheet is saved
    // really only use here is just to stop saving copies
    pub path:       String,

    // NOTE amount of Horizontal Cuts, a spritesheet with 5 frames of animation will have 5 horizontal cuts
    pub cuts_x:     u32,

    // NOTE possibly making it check vertically for saving multiple animations in one Texture?
    //cuts_y:   u32,

    // NOTE maybe add an offset for each sprite
    // offsets: Vec<cgmath::Vec2<f32>>

    // NOTE Hold frames, if 0 it means it will be displayed for only 1 frame, 1 means it will be held for 2 etc
    pub hold:    Vec<u32>
}

impl SpriteSheet
{
    /// creates a new Spritesheet
    pub fn new(path: &str, gl: &glow::Context, cuts_x: u32, hold: Vec<u32>) -> Self
    {
        let texture_id = texture_from_file(path, gl);

        SpriteSheet
        {
            texture_id,
            path: String::from(path),
            cuts_x,
            hold
        }
    }
}


/// Animation Component
///
/// currently it's using a hashmap with String keys
///
/// current also uses a String to identify the spritesheet
#[derive(Component)]
pub struct Animator
{
    // NOTE sprite sheets in a hashmap
    pub spritesheets: HashMap<AnimationType, SpriteSheet>,

    // NOTE current is possibly unneeded as it would be the job of a different System/Component to manage that
    pub current: AnimationType,

    // NOTE have a shader option to apply a different shader
    //pub shader: Option<shader::Shader>

    pub frame_time: f32


}

impl Default for Animator
{
    fn default() -> Self
    {
        Animator {
            spritesheets: HashMap::new(),
            current: AnimationType::Idle,
            frame_time: 0.0,
        }
    }
}
// TODO should pre-jump and jump frames be seperate?
// WIP all the animations that *could* show up, most of the air ones are expected to exist
// However just because an AnimationType is in this list doesn't mean it has to exist or be loaded in
#[derive(PartialEq, Eq, Hash)]
pub enum AnimationType
{
    // Basic Animations, universal
    Idle,
    StandUp,
    CrouchDown, // <- this is the animation of crouching down after standing
    Crouch, // <- this is for the loop of being crouching
    WalkingForward,
    WalkingBackward,
    Guard,
    CrouchGuard,
    AirGuard,
    JumpNeutral,
    JumpBackward,
    JumpForward,
    Fall,
    // Extra Movement animations, not universal
    AirDash,
    AirBackDash,
    Running,
    BackDashing,
    // Normals, Universal
    StandLP,
    StandMP,
    StandHP,
    StandLK,
    StandMK,
    StandHK,
    CrouchLP,
    CrouchMP,
    CrouchHP,
    CrouchLK,
    CrouchMK,
    CrouchHK,
    AirLP,
    AirMP,
    AirHP,
    AirLK,
    AirMK,
    AirHK,
    // hit animations, expect more complicated ones to show up, Universal
    CrouchHit,
    StandHit,
    AirHit,
    // command Normals, maybe would be Better to have them in the Specials
    CommandNormal01,
    CommandNormal02,
    CommandNormal03,
    CommandNormal04,
    CommandNormal05,
    CommandNormal06,
    CommandNormal07,
    CommandNormal08,
    CommandNormal09,
    CommandNormal10,
    // Specials, any Miscallenous Animations will also end up here
    // This includes fireballs
    // Rekkas will be seperated into different Animations
    Special01,
    Special02,
    Special03,
    Special04,
    Special05,
    Special06,
    Special07,
    Special08,
    Special09,
    Special10,
    Special11,
    Special12,
    Special13,
    Special14,
    Special15,
    // Supers
    Super01,
    Super02,
    Super03,
    Super04,
    Super05,
}




/// load texture from a filepath and return the ID
///
/// NOTE this function does not check if the texture has already loaded
fn texture_from_file(path: &str, gl: &glow::Context) -> glow::NativeTexture
    {

        let texture_id;
        unsafe {
            texture_id = gl.create_texture().unwrap();

            let img = image::open(&Path::new(&path)).expect("texture failed to load");
            let img = img.flipv();

            let format = match img {
                image::DynamicImage::ImageLuma8(_)  => glow::RED,
                image::DynamicImage::ImageLumaA8(_) => glow::RG,
                image::DynamicImage::ImageRgb8(_)   => glow::RGB,
                image::DynamicImage::ImageRgba8(_)  => glow::RGBA,
                _ => panic!("unknown image format at {path}")
            };

            let data = img.as_bytes();

            gl.bind_texture(glow::TEXTURE_2D, Some(texture_id));
            gl.tex_image_2d(glow::TEXTURE_2D, 0, format as i32, img.width() as i32, img.height() as i32,
                            0, format, glow::UNSIGNED_BYTE, Some(&data));
            gl.generate_mipmap(glow::TEXTURE_2D);


            gl.tex_parameter_i32(glow::TEXTURE_2D, glow::TEXTURE_WRAP_S, glow::REPEAT as i32);
            gl.tex_parameter_i32(glow::TEXTURE_2D, glow::TEXTURE_MAG_FILTER, glow::NEAREST as i32);
            gl.tex_parameter_i32(glow::TEXTURE_2D, glow::TEXTURE_MAG_FILTER, glow::NEAREST_MIPMAP_NEAREST as i32);

        }

        texture_id
}
